# coding: utf-8
try:
    import os
    import tensorflow as tf
    from tensorflow.python.saved_model import signature_constants
    from tensorflow.python.saved_model import tag_constants
    from tensorflow.python.framework import convert_to_constants
    from tensorflow.python.framework import tensor_shape
    from tensorflow.python.saved_model import save
except:
    pass


def get_name(tensor_name):
    """
    去掉模型名字中的:0
    :param tensor_name:
    :return:
    """
    return "".join(tensor_name.split(":")[:-1])


def read_pb_model(pb_path):
    """
    读Pb模型
    :param pb_path:
    :return:
    """
    with tf.gfile.GFile(pb_path, "rb") as f:
        graph_def = tf.GraphDef()
        graph_def.ParseFromString(f.read())
        return graph_def


def covert_pb_saved_model(graph_def, export_dir, input_list, output_list):
    """
    提供pb转写saved_model格式
    :param graph_def:
    :param export_dir:
    :param input_list:
    :param output_list:
    :return:
    """
    # 如果输入是空的，解析node提示可能的输入输出
    if (not input_list) and (not output_list):
        guess_input, guess_output = [], []
        for node_index, node in enumerate(list(graph_def.node)):
            if node.op.lower() == "placeholder":
                guess_input.append(node.name)
                continue
            if node_index >= len(graph_def.node) - 5:
                guess_output.append(node.name)
        print(f"guess input: {guess_input}")
        print(f"guess output: {guess_output}")
        input_str = input("输入：（多个以空格分开，需要带上:0）")
        input_list = input_str.split()
        print(f"input_list: {input_list}")
        output_str = input("输出：（多个以空格分开，需要带上:0）")
        output_list = output_str.split()
        print(f"output_list: {output_list}")
    builder = tf.saved_model.builder.SavedModelBuilder(export_dir)
    sigs = {}
    input_dict, output_dict = {}, {}
    with tf.Session(graph=tf.Graph()) as sess:
        tf.import_graph_def(graph_def, name="")
        g = tf.get_default_graph()
        for input_name in input_list:
            input_dict[get_name(input_name)] = g.get_tensor_by_name(input_name)
        for output_name in output_list:
            output_dict[get_name(output_name)] = g.get_tensor_by_name(output_name)
        sigs[signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY] = \
            tf.saved_model.signature_def_utils.predict_signature_def(
                input_dict, output_dict)
        builder.add_meta_graph_and_variables(sess,
                                             [tag_constants.SERVING],
                                             signature_def_map=sigs)
        builder.save()


def covert_pb_to_server_model(pb_path, export_dir, input_list=None, output_list=None):
    """
    转换pb到saved_model
    :param pb_path:
    :param export_dir:
    :param input_list: 模型的输入
    :param output_list: 模型的输出
    :return:
    """
    if os.path.exists(export_dir):
        print("export_dir is already exists, please remove")
        return
    graph_def = read_pb_model(pb_path)
    covert_pb_saved_model(graph_def, export_dir, input_list, output_list)


if __name__ == '__main__':
    dirname = "/Users/lhy/Desktop/pmu_test/202204145"
    pb_model_path = os.path.join(dirname, "model.pb")
    saved_model_path = os.path.join(dirname, "model.savedmodel")
    covert_pb_to_server_model(pb_model_path,
                              saved_model_path,
                              [],
                              [])



# coding: utf-8
import math
import subprocess
import time
from collections import Counter
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
import copy

import numpy as np


def read_gpu(return_least=False):
    import pynvml
    pynvml.nvmlInit()
    device_count = pynvml.nvmlDeviceGetCount()
    gpu_list = []
    for i in range(device_count):
        handle = pynvml.nvmlDeviceGetHandleByIndex(i)
        meminfo = pynvml.nvmlDeviceGetMemoryInfo(handle)
        name = pynvml.nvmlDeviceGetName(handle).decode()
        total = round(meminfo.total / 1024 ** 3, 2)
        used = round(meminfo.used / 1024 ** 3, 2)
        free = round(meminfo.free / 1024 ** 3, 2)
        information = {
            "id": i,
            "name": name,
            "total": total,
            "used": used,
            "free": free
        }
        gpu_list.append(information)
    if return_least:
        sort_gpu_list = copy.copy(gpu_list)
        sort_gpu_list.sort(key=lambda x: (x["used"], -x["id"]))
        least_id = sort_gpu_list[0]["id"]
        return gpu_list, least_id
    else:
        return gpu_list


def cut_list(target, batch_size):
    return [target[i:i + batch_size] for i in range(0, len(target), batch_size)]


def dict_set_value(input_data, args):
    assert len(args) == len(input_data.keys())
    for i, k in enumerate(input_data.keys()):
        input_data[k].append(args[i])


def l2_normalize(vecs):
    """l2标准化
    :param vecs: np.ndarray
    """
    norms = (vecs ** 2).sum(axis=1, keepdims=True) ** 0.5
    return vecs / np.clip(norms, 1e-8, np.inf)


def data_count(text_list, level="char"):
    """
    统计一个list的文本的长度
    """
    assert level in ["char", "word", "c", "w"]
    count_list = []
    for text in text_list:
        token_num = len(text.split()) if level[0] == "w" else len(text)
        count_list.append(token_num)
    counter = Counter(count_list)
    high_freq = counter.most_common(1)[0]
    result = {
        "min_length": min(count_list),
        "max_length": max(count_list),
        "ave_length": int(sum(count_list) / len(count_list)),
        "high_freq_length": high_freq[0],
        "high_freq_numbers": high_freq[1],
        "counter": counter
    }
    return result


def exec_shell(cmd):
    """打印并执行命令内容，并支持写入日志文件中
    Args:
        cmd: 执行命令的内容（str）
    Returns:
        status: 执行状态
    """
    print(cmd)
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=1)
    while p.poll() is None:
        line = p.stdout.readline().strip().decode()
        if line:
            print(line)
    status = p.returncode
    if status != 0:
        print(f'exec cmd failed. {cmd}')
    return status


class MultiProcessBase:
    """
    class MT(MultiProcessBase):
        @staticmethod
        def task(inputs):
            return [i ** 8 for i in inputs]
    m=MT().run()
    """

    def __init__(self, data, work_nums=4, batch_size=None, pool_type="t"):
        if batch_size:
            batch_size = batch_size
        else:
            batch_size = math.ceil(len(data) / work_nums)
        self.input_list = cut_list(data, batch_size)  # 每个进程的数据
        if pool_type == "t":
            self.pool = ThreadPoolExecutor(work_nums)
        else:
            self.pool = ProcessPoolExecutor(work_nums)

    @staticmethod
    def task(inputs):
        raise NotImplemented

    def run(self):
        obj_list = []
        for p_id in range(len(self.input_list)):
            r = self.pool.submit(self.task, self.input_list[p_id])
            obj_list.append(r)
        self.pool.shutdown()
        result_list = []
        for obj in obj_list:
            result_list.extend(obj.result())
        return result_list


class MultiThreadBar:
    """
    for i in MultiThreadBar(data, "te"):
        pass
    """

    def __init__(self, iter_data, desc=None):
        if not hasattr(iter_data, "__iter__"):
            raise ValueError("data必须是迭代器")
        self.data = iter_data
        self.length = len(iter_data)
        self.index = 0
        self.start_time = time.time()
        self.cost_time = 0
        self.time_step = 0
        self.last_time = 0
        self.msg = "\r"
        if desc:
            self.msg = self.msg + desc + ": "

    def update_time_step(self):
        self.cost_time = time.time() - self.start_time
        self.time_step = self.cost_time / self.index
        self.last_time = self.time_step * (self.length - self.index)

    def __iter__(self):
        return self

    def __next__(self):
        if self.index != 0:
            self.update_time_step()
            print(f"{self.msg}{self.index} / {self.length - 1}, time: {int(self.cost_time)}s / {int(self.last_time)}s",
                  end="")
        else:
            print(f"{self.msg}{self.index} / {self.length - 1}", end="")
        if self.index == self.length:
            print(f"{self.msg}{self.index - 1} / {self.length - 1}, cost: {int(self.cost_time)}")
            raise StopIteration
        value = self.data[self.index]
        self.index += 1
        return value


def write_excel(path: str,
                write_data_list: list,
                sheet_name_list: list = None,
                format_title: dict = None,
                format_content: dict = None,
                row_height: list = None,
                col_width: list = None):
    """

    :param path: 文件路径
    :param write_data_list: 文件内容, 几张表就几个data，data里面的结构是{"表头": [内容1,内容2,...]}
    :param sheet_name_list:
    :param format_title:
    :param format_content:
    :param row_height:
    :param col_width:
    :return:
    """
    import xlsxwriter
    # 构建A-Z到0-25的映射
    char2num = {chr(i): i - ord("A") for i in range(ord("A"), ord("Z") + 1)}
    char2num.update({chr(i): i - ord("a") for i in range(ord("a"), ord("z") + 1)})
    # 构建format
    format_t = {"bold": True, "align": "center", "valign": "vcenter", "text_wrap": True}
    format_c = {"align": "center", "valign": "vcenter", "text_wrap": True}
    if format_title is not None:
        format_t.update(format_title)
    if format_content is not None:
        format_c.update(format_content)
    # 构建sheet_name
    sheet_num = len(write_data_list)
    if sheet_name_list is None:
        sheet_name_list = [f"sheet{i}" for i in range(sheet_num)]
    assert sheet_num == len(sheet_name_list)
    # 构建数据
    workbook = xlsxwriter.Workbook(path)
    format_t = workbook.add_format(format_t)
    format_c = workbook.add_format(format_c)
    for data, sheet_name in zip(write_data_list, sheet_name_list):
        worksheet = workbook.add_worksheet(sheet_name)
        # 写行高，如果有
        if row_height is not None:
            for row, height in row_height:
                worksheet.set_row(row, height)
        # 写列宽，如果有
        if col_width is not None:
            for col, width in col_width:
                col = char2num.get(col, col)
                worksheet.set_column(col, col, width)
        title_list = list(data.keys())
        for title_index, title in enumerate(title_list):
            worksheet.write(0, title_index, title, format_t)
            for item_index, item in enumerate(data[title]):
                worksheet.write(item_index + 1, title_index, item, format_c)
    workbook.close()


class ApiTestBase:
    def __init__(self, url=None, port=None, wait_time=300, batch_size=1, work_num=1,
                 request_type="g", cpu_num=None, logger=None):
        self.url = url
        self.port = port
        self.wait_time = wait_time  # 压测时间
        self.batch_size = batch_size
        self.work_num = work_num  # 并发请求数
        self.request_type = request_type  # 请求类型，g(grpc)或者h(http)
        self.cpu_num = cpu_num  # 服务端是否限制了核数，用于统计qps/c
        self.print = print if logger is None else logger.info
        self.pool = ThreadPoolExecutor(self.work_num)  # 压测线程池
        self.success_list = list()  # 成功次数
        self.failure_list = list()  # 失败次数

    def init_client(self):
        """
        初始化请求客户端
        :return:
        """
        raise NotImplemented

    def task(self):
        raise NotImplemented

    def main(self):
        self.print("task start")
        task_list = []
        for _ in range(self.work_num):
            t = self.pool.submit(self.task)
            task_list.append(t)
        self.pool.shutdown()
        for t in task_list:
            success_times, failure_times = t.result()
            self.success_list.append(success_times)
            self.failure_list.append(failure_times)
        self.report()

    def report(self):
        """
        获得统计结果
        :return:
        """
        total_success_times = sum(self.success_list)
        total_failure_times = sum(self.failure_list)
        failure_rate = round(total_failure_times / (total_success_times + total_failure_times), 3)
        qps = round(total_success_times / self.wait_time, 2)
        self.print(f"qps: {qps}")
        self.print(f"failure rate: {failure_rate}")
        if self.cpu_num is not None:
            qps_c = round(qps / self.cpu_num, 2)
            self.print(f"qps/c: {qps_c}")



# coding: utf-8
import math
from collections import Counter

import numpy as np
from .lhy_tool_utils import l2_normalize


def cal_core_pagerank(input_array):
    from DaPy.methods import PageRank
    """
    计算新的簇心
    """
    similar_mat = np.dot(input_array, input_array.T)
    similar_mat /= similar_mat.sum(axis=0)
    pg = PageRank(engine="numpy")
    node_weight = pg.transform(similar_mat)
    scores = np.argsort(-np.array(node_weight))  # 根据pagerank的重要性求出簇心群
    candidate_length = math.ceil(len(input_array) / 10)  # 取前10%个作为簇心
    scores_ids = scores[:candidate_length]
    new_core = np.mean(input_array[scores_ids], axis=0)  # 计算簇心群的平均向量作为新簇心
    return new_core.reshape(1, -1)


def single_pass(train_array: np.ndarray, text_list=None, alpha=0.8, beta=0.9, theta=0.95, core_init=50):
    """
    1、先对簇心进行初始化
        1.1、对聚类的文本进行统计，找到重复度最大的样本，作为第一个簇心（如果没有给文本列表就使用第一个样本作为第一个簇心）
        1.2、找到所有样本中，与第一个簇心相似度最低的簇心，作为第二个簇心
        1.3、遍历剩余样本，计算每个样本与当前簇心的相似度的均值，与均值最低且不大于0.7的样本作为下一个簇心
        1.4、当簇心数量大于等于预设值，或者，在每一次遍历过程中没有发生簇心的新增，即判断簇心的初始化已完成
    2、根据相似度决定把剩余样本放入簇还是新建簇
        2.1、计算样本与现有簇的相似度，取最大值，如果大于等于alpha放入簇，且如果大于等于alpha+0.5更新簇心
        2.1、否则新增一个簇
    3、对所有的簇进行合并
        3.1、计算所有簇之间的相似度矩阵，并mask掉下三角
        3.2、将所有的相似度高->低进行排序，优先合并高相似度的，相似度小于beta的不做合并
    4、簇内排序，根据样本到簇心的相似度
    5、簇间排序，根据簇的大小
    :param train_array: 要聚类的向量集合
    :param text_list: 可给可不给，如果给了，将text重复次数最多的作为第一个簇心
    :param alpha: 样本相似度阈值
    :param beta: 簇相似度阈值
    :param theta: 需要调整簇心的相似度
    :param core_init: 初始多少个簇心
    :return cluster_result_list: ["cluster_id": int, "items": [], "core": np.ndarray(可选)]
    """
    assert len(train_array.shape) == 2
    # 初始化最终结果
    cluster_result_list = []  # 存放最终返回的结果
    # 1、对簇心进行初始化
    data_num, feature_dim = train_array.shape
    if text_list is not None:
        most_text = Counter(text_list).most_common(1)[0][0]  # 出现次数最多的text
        first_init_index = text_list.index(most_text)  # 出现次数最多的text的第一个下标
    else:
        first_init_index = 0
    core_vec = None  # 初始簇心
    visited_index = set()  # 存储已被使用过的样本
    while True:
        add_bz = False  # 是否新增簇心了
        if len(cluster_result_list) == 0:  # 如果簇心暂时为空，取出现次数最大的text
            visited_index.add(first_init_index)
            cluster_template = {"cluster_id": len(cluster_result_list), "items": [first_init_index],
                                "core": train_array[first_init_index].reshape(1, -1)}
            cluster_result_list.append(cluster_template)
            core_vec = train_array[first_init_index].reshape(1, -1)
            add_bz = True
        else:
            tmp_similar = np.dot(core_vec, train_array.T)
            similar_ave = np.mean(tmp_similar, axis=0)
            similar_sort_idx = np.argsort(similar_ave)
            for can_idx in similar_sort_idx:
                if similar_ave[can_idx] >= alpha:
                    break
                if can_idx not in visited_index and np.max(tmp_similar[:, can_idx]) < beta:
                    visited_index.add(can_idx)
                    cluster_template = {"cluster_id": len(cluster_result_list), "items": [can_idx],
                                        "core": train_array[can_idx].reshape(1, -1)}
                    cluster_result_list.append(cluster_template)
                    core_vec = np.append(core_vec, train_array[can_idx].reshape(1, -1), axis=0)
                    add_bz = True
                    break
        if (not add_bz) or len(cluster_result_list) >= core_init:  # 如果某一次循环没有新增簇心，或者簇数已达到一定数量
            break
    # 2、根据相似度决定把剩余样本放入簇还是新建簇
    for sample_index, sample_array in enumerate(train_array):
        if sample_index not in visited_index:
            cluster_similar = np.dot(sample_array, core_vec.T)  # 计算该样本与现有簇心的相似度，不能一次性计算好
            max_cluster_idx = np.argmax(cluster_similar)  # 最大相似度的簇
            max_similarity = cluster_similar[max_cluster_idx]  # 最大相似度
            if max_similarity >= alpha:  # 最大相似度大于等于阈值，则合并
                cluster_result_list[max_cluster_idx]["items"].append(sample_index)
                if max_similarity >= theta:
                    new_core = np.mean([sample_array.reshape(1, -1), core_vec[max_cluster_idx]], axis=0)
                    new_core = l2_normalize(new_core)
                    core_vec[max_cluster_idx] = new_core
                    cluster_result_list[max_cluster_idx]["core"] = new_core
            else:  # 相似度不够就新增一个簇
                cluster_template = {"cluster_id": len(cluster_result_list), "items": [sample_index],
                                    "core": sample_array.reshape(1, -1)}
                cluster_result_list.append(cluster_template)
                core_vec = np.append(core_vec, sample_array.reshape(1, -1), axis=0)
    # 3、对所有的簇进行合并， core_vec 和 cluster_result_list是一一对应的
    combine_cluster_result = []  # 重置一下合并过的簇
    combine_core = None
    while True:
        combine_bz = False  # 该次循环是否产生了合并
        if combine_cluster_result and combine_core is not None:
            cluster_result_list = combine_cluster_result
            core_vec = combine_core
            combine_cluster_result = []  # 重置一下合并过的簇
            combine_core = None
        cluster_core_similarity = np.dot(core_vec, core_vec.T)
        cluster_core_similarity = np.triu(cluster_core_similarity, k=1)  # 剔除下三角
        cluster_num = len(cluster_core_similarity)
        sim_argsort = np.argsort(-cluster_core_similarity.reshape(-1, ))
        for index in sim_argsort:
            row, col = divmod(index, cluster_num)
            if cluster_core_similarity[row][col] >= beta:
                if len(cluster_result_list[col]["items"]) > 0:
                    combine_bz = True
                    # 把列的簇中所有数据放入行的簇
                    cluster_result_list[row]["items"].extend(cluster_result_list[col]["items"])
                    cluster_result_list[row]["core"] = None
                    cluster_result_list[col]["items"] = []
            else:
                break
        for cluster in cluster_result_list:
            cluster_length = len(cluster["items"])
            if cluster_length > 0:
                cluster["cluster_id"] = len(combine_cluster_result)
                if cluster["core"] is None:
                    cluster_vector = train_array[cluster["items"]]
                    core = cal_core_pagerank(cluster_vector)
                    core = l2_normalize(core)
                else:
                    core = cluster["core"]
                combine_cluster_result.append(cluster)
                combine_core = core if combine_core is None else np.append(combine_core, core, axis=0)
        if not combine_bz:
            break
    # 簇内排序, 利用cluster_result_list和core_vec进行排序
    cluster_num = len(cluster_result_list)
    assert len(core_vec) == cluster_num
    for c_index in range(cluster_num):
        cluster = cluster_result_list[c_index]
        cluster_core = core_vec[c_index]
        cluster["core"] = cluster_core
        if len(cluster["items"]) > 1:
            cluster_in_similar = np.dot(cluster_core, train_array[cluster["items"]].T)
            sorted_index_list = np.argsort(-cluster_in_similar)
            cluster["items"] = [cluster["items"][i] for i in sorted_index_list]
    # 簇间排序
    cluster_result_list.sort(key=lambda x: len(x["items"]), reverse=True)
    for c_index, cluster in enumerate(cluster_result_list):  # 重置簇的id
        cluster["cluster_id"] = c_index
    # 验证
    return cluster_result_list
